//
//  HomeViewController.swift
//  Reader
//
//  Created by Shadowfacts on 11/25/21.
//

import UIKit
import CoreData

protocol HomeViewControllerDelegate: AnyObject {
    func switchToAccount(_ account: LocalData.Account)
}

class HomeViewController: UIViewController {
    
    weak var delegate: HomeViewControllerDelegate?
    
    let fervorController: FervorController
    
    private var collectionView: UICollectionView!
    private var dataSource: UICollectionViewDiffableDataSource<Section, Item>!
    private var groupResultsController: NSFetchedResultsController<Group>!
    private var feedResultsController: NSFetchedResultsController<Feed>!
    
    init(fervorController: FervorController) {
        self.fervorController = fervorController
        
        super.init(nibName: nil, bundle: nil)
    }
    
    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        // todo: account info
        title = "Reader"
        
        view.addInteraction(StretchyMenuInteraction(delegate: self))
        
        view.backgroundColor = .appBackground
        
        var config = UICollectionLayoutListConfiguration(appearance: .grouped)
        config.headerMode = .supplementary
        config.backgroundColor = .appBackground
        config.separatorConfiguration.topSeparatorVisibility = .visible
        config.separatorConfiguration.bottomSeparatorVisibility = .hidden
        config.itemSeparatorHandler = { indexPath, defaultConfig in
            var config = defaultConfig
            if indexPath.section == 0 && indexPath.row == 0 {
                config.topSeparatorVisibility = .hidden
            }
            return config
        }
        let layout = UICollectionViewCompositionalLayout.list(using: config)
        collectionView = UICollectionView(frame: view.bounds, collectionViewLayout: layout)
        collectionView.autoresizingMask = [.flexibleWidth, .flexibleHeight]
        collectionView.delegate = self
        view.addSubview(collectionView)
        
        dataSource = createDataSource()
        
        var snapshot = NSDiffableDataSourceSnapshot<Section, Item>()
        snapshot.appendSections([.all, .groups, .feeds])
        snapshot.appendItems([.unread, .all], toSection: .all)
        dataSource.apply(snapshot, animatingDifferences: false)
        
        let groupReq = Group.fetchRequest()
        groupReq.sortDescriptors = [NSSortDescriptor(key: "title", ascending: true)]
        groupResultsController = NSFetchedResultsController(fetchRequest: groupReq, managedObjectContext: fervorController.persistentContainer.viewContext, sectionNameKeyPath: nil, cacheName: nil)
        groupResultsController.delegate = self
        try! groupResultsController.performFetch()
        
        let feedReq = Feed.fetchRequest()
        feedReq.sortDescriptors = [NSSortDescriptor(key: "title", ascending: true)]
        feedResultsController = NSFetchedResultsController(fetchRequest: feedReq, managedObjectContext: fervorController.persistentContainer.viewContext, sectionNameKeyPath: nil, cacheName: nil)
        feedResultsController.delegate = self
        try! feedResultsController.performFetch()
    }
    
    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        
        if let indexPaths = collectionView.indexPathsForSelectedItems {
            for indexPath in indexPaths {
                collectionView.deselectItem(at: indexPath, animated: true)
            }
        }
        
        var snapshot = dataSource.snapshot()
        // reconfigure so that unread counts update
        snapshot.reconfigureItems(snapshot.itemIdentifiers)
        dataSource.apply(snapshot)
    }
    
    private func createDataSource() -> UICollectionViewDiffableDataSource<Section, Item> {
        let sectionHeaderCell = UICollectionView.SupplementaryRegistration<UICollectionViewListCell>(elementKind: UICollectionView.elementKindSectionHeader) { supplementaryView, elementKind, indexPath in
            let section = self.dataSource.sectionIdentifier(for: indexPath.section)!
            var config = supplementaryView.defaultContentConfiguration()
            config.text = section.title
            supplementaryView.contentConfiguration = config
        }
        let listCell = UICollectionView.CellRegistration<HomeCollectionViewCell, Item> { cell, indexPath, item in
            var config = UIListContentConfiguration.valueCell()
            config.text = item.title
            if let req = item.countFetchRequest,
               let count = try? self.fervorController.persistentContainer.viewContext.count(for: req) {
                config.secondaryText = "\(count)"
                config.secondaryTextProperties.color = .tintColor
            }
            cell.contentConfiguration = config
            
            cell.accessories = [.disclosureIndicator()]
        }
        let dataSource = UICollectionViewDiffableDataSource<Section, Item>(collectionView: collectionView) { collectionView, indexPath, item in
            return collectionView.dequeueConfiguredReusableCell(using: listCell, for: indexPath, item: item)
        }
        dataSource.supplementaryViewProvider = { collectionView, elementKind, indexPath in
            if elementKind == UICollectionView.elementKindSectionHeader {
                return collectionView.dequeueConfiguredReusableSupplementary(using: sectionHeaderCell, for: indexPath)
            } else {
                return nil
            }
        }
        return dataSource
    }
    
}

extension HomeViewController {
    enum Section: Hashable {
        case all
        case groups
        case feeds
        
        var title: String {
            switch self {
            case .all:
                return ""
            case .groups:
                return "Groups"
            case .feeds:
                return "Feeds"
            }
        }
    }
    enum Item: Hashable {
        case unread
        case all
        case group(Group)
        case feed(Feed)
        
        var title: String {
            switch self {
            case .unread:
                return "Unread Articles"
            case .all:
                return "All Articles"
            case let .group(group):
                return group.title
            case let .feed(feed):
                return feed.title!
            }
        }
        
        var fetchRequest: NSFetchRequest<Reader.Item> {
            let req = Reader.Item.fetchRequest()
            switch self {
            case .unread:
                req.predicate = NSPredicate(format: "read = NO")
            case .all:
                break
            case .group(let group):
                req.predicate = NSPredicate(format: "feed in %@", group.feeds!)
            case .feed(let feed):
                req.predicate = NSPredicate(format: "feed = %@", feed)
            }
            return req
        }
        
        var countFetchRequest: NSFetchRequest<Reader.Item>? {
            let req = Reader.Item.fetchRequest()
            switch self {
            case .unread:
                req.predicate = NSPredicate(format: "read = NO")
            case .all:
                return nil
            case .group(let group):
                req.predicate = NSPredicate(format: "read = NO AND feed in %@", group.feeds!)
            case .feed(let feed):
                req.predicate = NSPredicate(format: "read = NO AND feed = %@", feed)
            }
            return req
        }
    }
}

extension HomeViewController: NSFetchedResultsControllerDelegate {
    func controller(_ controller: NSFetchedResultsController<NSFetchRequestResult>, didChangeContentWith snapshot: NSDiffableDataSourceSnapshotReference) {
        var snapshot = dataSource.snapshot()
        if controller == groupResultsController {
            if snapshot.sectionIdentifiers.contains(.groups) {
                snapshot.deleteItems(snapshot.itemIdentifiers(inSection: .groups))
            }
            snapshot.appendItems(controller.fetchedObjects!.map { .group($0 as! Group) }, toSection: .groups)
        } else if controller == feedResultsController {
            if snapshot.sectionIdentifiers.contains(.feeds) {
                snapshot.deleteItems(snapshot.itemIdentifiers(inSection: .feeds))
            }
            snapshot.appendItems(controller.fetchedObjects!.map { .feed($0 as! Feed) }, toSection: .feeds)
        }
        dataSource.apply(snapshot, animatingDifferences: false)
    }
}

extension HomeViewController: UICollectionViewDelegate {
    func collectionView(_ collectionView: UICollectionView, didSelectItemAt indexPath: IndexPath) {
        guard let item = dataSource.itemIdentifier(for: indexPath) else {
            return
        }
        let vc = ItemsViewController(fetchRequest: item.fetchRequest, fervorController: fervorController)
        vc.title = item.title
        show(vc, sender: nil)
        UISelectionFeedbackGenerator().selectionChanged()
    }
    
    func collectionView(_ collectionView: UICollectionView, contextMenuConfigurationForItemAt indexPath: IndexPath, point: CGPoint) -> UIContextMenuConfiguration? {
        guard let item = dataSource.itemIdentifier(for: indexPath) else {
            return nil
        }
        return UIContextMenuConfiguration(identifier: nil, previewProvider: {
            return ItemsViewController(fetchRequest: item.fetchRequest, fervorController: self.fervorController)
        }, actionProvider: nil)
    }
    
    func collectionView(_ collectionView: UICollectionView, willPerformPreviewActionForMenuWith configuration: UIContextMenuConfiguration, animator: UIContextMenuInteractionCommitAnimating) {
        guard let vc = animator.previewViewController else {
            return
        }
        animator.preferredCommitStyle = .pop
        animator.addCompletion {
            self.show(vc, sender: nil)
        }
    }
}

extension HomeViewController: StretchyMenuInteractionDelegate {
    func stretchyMenuTitle() -> String? {
        return "Switch Accounts"
    }
    func stretchyMenuItems() -> [StretchyMenuItem] {
        var items: [StretchyMenuItem] = LocalData.accounts.map { account in
            var title = account.instanceURL.host!
            if let port = account.instanceURL.port, port != 80 && port != 443 {
                title += ":\(port)"
            }
            let subtitle = account.id == fervorController.account!.id ? "Currently logged in" : nil
            return StretchyMenuItem(title: title, subtitle: subtitle) { [unowned self] in
                guard account.id != self.fervorController.account!.id else { return }
                self.delegate?.switchToAccount(account)
            }
        }
        items.append(StretchyMenuItem(title: "Add Account", subtitle: nil, action: { [unowned self] in
            let login = LoginViewController()
            login.delegate = self
            login.navigationItem.leftBarButtonItem = UIBarButtonItem(systemItem: .cancel, primaryAction: UIAction(handler: { (_) in
                self.dismiss(animated: true)
            }), menu: nil)
            let nav = UINavigationController(rootViewController: login)
            self.present(nav, animated: true)
        }))
        return items
    }
}

extension HomeViewController: LoginViewControllerDelegate {
    func didLogin(with controller: FervorController) {
        self.dismiss(animated: true)
        (view.window!.windowScene!.delegate as! SceneDelegate).didLogin(with: controller)
    }
}
